"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""

from abc import ABC, abstractmethod
from logging import getLogger
from typing import Set

from sc_kpm.sc_agent import ScAgentAbstract


class ScModuleAbstract(ABC):
    """
    ScModule is a container for handling multiple ScAgent objects.
    You can add and remove agents while module is registered or unregistered.
    """

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def add_agent(self, agent: ScAgentAbstract) -> None:
        """Add agent to the module and register it if module is registered"""

    @abstractmethod
    def remove_agent(self, agent: ScAgentAbstract) -> None:
        """Remove agent from the module and unregister it if module is registered"""

    @abstractmethod
    def _register(self) -> None:
        """Register all agents in the module"""

    @abstractmethod
    def _unregister(self) -> None:
        """Unregister all agents from the module"""


class ScModule(ScModuleAbstract):
    def __init__(self, *agents: ScAgentAbstract) -> None:
        self._agents: Set[ScAgentAbstract] = {*agents}
        self._is_registered: bool = False
        self.logger = getLogger(f"{self.__module__}.{self.__class__.__name__}")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(map(repr, self._agents))})"

    def add_agent(self, agent: ScAgentAbstract) -> None:
        if self._is_registered:
            agent._register()  # pylint: disable=protected-access
        self._agents.add(agent)

    def remove_agent(self, agent: ScAgentAbstract) -> None:
        if self._is_registered:
            agent._unregister()  # pylint: disable=protected-access
        self._agents.remove(agent)

    def _register(self) -> None:
        if self._is_registered:
            self.logger.warning("Already registered")
            return
        if not self._agents:
            self.logger.warning("No agents to register")
        for agent in self._agents:
            agent._register()  # pylint: disable=protected-access
        self._is_registered = True
        self.logger.info("Registered")

    def _unregister(self) -> None:
        for agent in self._agents:
            agent._unregister()  # pylint: disable=protected-access
        self._is_registered = False
        self.logger.info("Unregistered")



"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""
import os
import signal
import threading

from sc_client.constants.common import ScEventType
from sc_client.models import ScAddr

from sc_kpm import ScAgent, ScAgentClassic, ScModule, ScResult
from sc_kpm.identifiers import CommonIdentifiers
from sc_kpm.utils.action_utils import execute_agent, finish_action_with_status
from tests.common_tests import BaseTestCase

WAIT_TIME = 1


class CommonTests(BaseTestCase):
    def test_sc_agents(self):
        class Agent(ScAgent):
            ACTION_CLASS_NAME = "test_agent"

            def __init__(self):
                super().__init__(self.ACTION_CLASS_NAME, ScEventType.ADD_OUTGOING_EDGE)

            def on_event(self, event_element: ScAddr, event_edge: ScAddr, action_element: ScAddr) -> ScResult:
                finish_action_with_status(action_element, True)
                return ScResult.OK

        class AgentClassic(ScAgentClassic):
            ACTION_CLASS_NAME = "test_agent_classic"

            def __init__(self):
                super().__init__(self.ACTION_CLASS_NAME)

            def on_event(self, event_element: ScAddr, event_edge: ScAddr, action_element: ScAddr) -> ScResult:
                finish_action_with_status(action_element, True)
                return ScResult.OK

        module = ScModule(Agent(), AgentClassic())
        self.server.add_modules(module)
        kwargs = dict(
            arguments={},
            concepts=[],
            initiation=Agent.ACTION_CLASS_NAME,
            wait_time=WAIT_TIME,
        )
        kwargs_classic = dict(
            arguments={},
            concepts=[CommonIdentifiers.QUESTION, AgentClassic.ACTION_CLASS_NAME],
            wait_time=WAIT_TIME,
        )
        self.assertFalse(execute_agent(**kwargs)[1])
        self.assertFalse(execute_agent(**kwargs_classic)[1])
        with self.server.register_modules():
            self.assertTrue(execute_agent(**kwargs)[1])
            self.assertTrue(execute_agent(**kwargs_classic)[1])
        self.assertFalse(execute_agent(**kwargs)[1])
        self.assertFalse(execute_agent(**kwargs_classic)[1])
        self.server.remove_modules(module)

    def test_sc_module(self):
        class TestAgent(ScAgent):
            def on_event(self, event_element: ScAddr, event_edge: ScAddr, action_element: ScAddr) -> ScResult:
                finish_action_with_status(action_element, True)
                return ScResult.OK

        def is_executing_successful(i: int) -> bool:
            return execute_agent(
                arguments={},
                concepts=[],
                initiation=f"agent{i}",
                wait_time=WAIT_TIME,
            )[1]

        agent1 = TestAgent("agent1", ScEventType.ADD_OUTGOING_EDGE)
        agent2 = TestAgent("agent2", ScEventType.ADD_OUTGOING_EDGE)
        agent3 = TestAgent("agent3", ScEventType.ADD_OUTGOING_EDGE)

        module1 = ScModule(agent1)
        module2 = ScModule(agent2)
        self.server.add_modules(module1)
        module1.add_agent(agent3)
        with self.server.register_modules():
            self.assertTrue(is_executing_successful(1))
            self.assertFalse(is_executing_successful(2))
            self.assertTrue(is_executing_successful(3))

            self.server.add_modules(module2)
            self.assertTrue(is_executing_successful(2))

            module1.remove_agent(agent3)
            self.assertTrue(is_executing_successful(1))
            self.assertFalse(is_executing_successful(3))
            self.server.remove_modules(module1, module2)

    def test_sc_server(self):
        class TestAgent(ScAgent):
            ACTION_CLASS_NAME = "some_agent"

            def __init__(self):
                super().__init__(self.ACTION_CLASS_NAME, ScEventType.ADD_OUTGOING_EDGE)

            def on_event(self, event_element: ScAddr, event_edge: ScAddr, action_element: ScAddr) -> ScResult:
                finish_action_with_status(action_element, True)
                return ScResult.OK

        def is_executing_successful() -> bool:
            return execute_agent(
                arguments={},
                concepts=[],
                initiation=TestAgent.ACTION_CLASS_NAME,
                wait_time=WAIT_TIME,
            )[1]

        module = ScModule(TestAgent())
        self.server.add_modules(module)
        self.assertFalse(is_executing_successful())
        self.server.register_modules()
        self.assertTrue(is_executing_successful())
        self.server.unregister_modules()
        self.assertFalse(is_executing_successful())

        main_pid = os.getpid()

        def execute_and_send_sigint():
            self.assertTrue(is_executing_successful())
            os.kill(main_pid, signal.SIGINT)
            self.assertFalse(is_executing_successful())

        with self.server.register_modules():
            thread = threading.Thread(target=execute_and_send_sigint, daemon=True)
            thread.start()
            self.server.serve()

        self.server.remove_modules(module)



from __future__ import annotations

from typing import Iterator

from sc_client.client import create_elements, delete_elements, template_search
from sc_client.constants import ScType, sc_types
from sc_client.models import ScAddr, ScConstruction, ScTemplate, ScTemplateResult

from sc_kpm.utils.common_utils import create_node


class ScSet:
    """
    ScSet is a class for handling set construction in kb.

    It has main set_node and edged elements.
    """

    def __init__(self, *elements: ScAddr, set_node: ScAddr = None, set_node_type: ScType = None) -> None:
        """
        Initialize ScSet.

        Receive or create set_node and add elements.

        :param elements: Elements of a set to initialize it.
        :param set_node: Set node has edges to each element.
        Optional parameter: it will be created if it doesn't exist.
        :param set_node_type: ScType for creating set node.
        """

        if set_node is None:
            if set_node_type is None:
                set_node_type = sc_types.NODE_CONST
            set_node = create_node(set_node_type)
        self._set_node = set_node
        self.add(*elements)

    def add(self, *elements: ScAddr) -> None:
        """Add elements to ScSet"""
        if elements:
            construction = ScConstruction()
            for element in elements:
                construction.create_edge(sc_types.EDGE_ACCESS_CONST_POS_PERM, self._set_node, element)
            create_elements(construction)

    @property
    def set_node(self) -> ScAddr:
        """Get the main element of ScSet"""
        return self._set_node

    def __eq__(self, other: ScSet) -> bool:
        return self._set_node == other._set_node

    @property
    def elements_set(self) -> set[ScAddr]:
        """Set of elements without order and duplicates"""
        search_results = self._elements_search_results()
        elements = {result[2] for result in search_results}
        return elements

    def __len__(self) -> int:
        """Get ScSet power"""
        return len(self.elements_set)  # No duplicates

    def __bool__(self) -> bool:
        """Check ScSet is not empty"""
        return bool(self._elements_search_results())

    def is_empty(self) -> bool:
        """Check if ScSet doesn't contain any element"""
        return not bool(self)

    def __iter__(self) -> Iterator[ScAddr]:
        """Iterate by ScSet elements"""
        return iter(self.elements_set)

    def __contains__(self, element: ScAddr) -> bool:
        """Check if ScSet contains element"""
        return element in self.elements_set

    def remove(self, *elements: ScAddr) -> None:
        """Remove the connections between set_node and elements"""
        templ = ScTemplate()
        for element in elements:
            templ.triple(self._set_node, sc_types.EDGE_ACCESS_VAR_POS_PERM, element)
        template_results = template_search(templ)
        delete_elements(*(res[1] for res in template_results))

    def clear(self) -> None:
        """Remove the connections between set_node and all elements"""
        template_results = self._elements_search_results()
        delete_elements(*(res[1] for res in template_results))

    def _elements_search_results(self) -> list[ScTemplateResult]:
        """Template search of all elements"""
        templ = ScTemplate()
        templ.triple(self._set_node, sc_types.EDGE_ACCESS_VAR_POS_PERM, sc_types.UNKNOWN)
        return template_search(templ)



from sc_kpm.sc_sets.sc_numbered_set import ScNumberedSet
from sc_kpm.sc_sets.sc_oriented_set import ScOrientedSet
from sc_kpm.sc_sets.sc_set import ScSet
from sc_kpm.sc_sets.sc_structure import ScStructure



"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""

from sc_kpm.utils import action_utils
from sc_kpm.utils.common_utils import (
    check_edge,
    create_binary_relation,
    create_edge,
    create_edges,
    create_link,
    create_links,
    create_node,
    create_nodes,
    create_norole_relation,
    create_role_relation,
    delete_edges,
    get_edge,
    get_edges,
    get_element_by_norole_relation,
    get_element_by_role_relation,
    get_link_content_data,
    get_system_idtf,
    search_role_relation_template,
)



from sc_client.client import check_elements
from sc_client.constants import ScType, sc_types
from sc_client.constants.exceptions import InvalidTypeError
from sc_client.models import ScAddr

from sc_kpm.sc_sets.sc_set import ScSet


class ScStructure(ScSet):
    """
    ScStructure is a class for handling structure construction in kb.

    It has main set_node with type NODE_CONST_STRUCT and edged elements.
    """

    def __init__(self, *elements: ScAddr, set_node: ScAddr = None, set_node_type: ScType = None) -> None:
        if set_node_type is None:
            set_node_type = sc_types.NODE_CONST_STRUCT
        if set_node is not None:
            set_node_type = check_elements(set_node)[0]
        if not set_node_type.is_struct():
            raise InvalidTypeError
        super().__init__(*elements, set_node=set_node, set_node_type=set_node_type)



from typing import Iterator, List, Optional

from sc_client.client import delete_elements, template_generate, template_search
from sc_client.constants import sc_types
from sc_client.models import ScAddr, ScTemplate, ScTemplateResult

from sc_kpm.identifiers import CommonIdentifiers, ScAlias
from sc_kpm.sc_keynodes import ScKeynodes
from sc_kpm.sc_sets.sc_set import ScSet
from sc_kpm.utils.common_utils import create_edge, create_role_relation, search_role_relation_template


class ScOrientedSet(ScSet):
    """
    ScOrientedSet is a class for handling oriented set structure in kb.

    It has main set_node and edged elements:
    Edge to the first element marked with 'rrel_1' node.
    The other have edges between edges from set_node marked with 'nrel_basic_sequence'.
    """

    def add(self, *elements: ScAddr) -> None:
        """Add elements to ScOrientedSet"""
        if elements:
            elements_iterator = iter(elements)
            current_edge = (
                self._create_first_element_edge(next(elements_iterator))
                if self.is_empty()
                else self._get_last_edge_and_delete_rrel_last()
            )
            for element in elements_iterator:
                current_edge = self._create_next_edge(current_edge, element)
            self._mark_edge_with_rrel_last(current_edge)

    def __iter__(self) -> Iterator[ScAddr]:
        """Iterate by ScOrientedSet elements"""
        start_template = search_role_relation_template(self._set_node, ScKeynodes[CommonIdentifiers.RREL_ONE])
        if not start_template:
            return
        yield start_template.get(ScAlias.ELEMENT)
        next_edge = start_template.get(ScAlias.ACCESS_EDGE)
        while True:
            elem_search_result = self._search_next_element_template(next_edge)
            if elem_search_result is None:
                return
            yield elem_search_result.get(ScAlias.ELEMENT)
            next_edge = elem_search_result.get(ScAlias.ACCESS_EDGE)

    @property
    def elements_list(self) -> List[ScAddr]:
        """List of elements with order"""
        return list(self)

    def remove(self, *elements: ScAddr) -> None:
        """Clear and add existing elements without given ones"""
        # TODO: optimize
        elements_new = [element for element in self.elements_list if element not in elements]
        self.clear()
        self.add(*elements_new)

    def _create_first_element_edge(self, element: ScAddr) -> ScAddr:
        """Create marked with rrel_1 edge to first element"""
        return create_role_relation(self._set_node, element, ScKeynodes[CommonIdentifiers.RREL_ONE])

    def _get_last_edge_and_delete_rrel_last(self) -> Optional[ScAddr]:
        """Search last edge of ScOrientedSet is it exists"""
        # Search marked last edge
        template = ScTemplate()
        template.triple_with_relation(
            self._set_node,
            sc_types.EDGE_ACCESS_VAR_POS_PERM >> ScAlias.ACCESS_EDGE,
            sc_types.UNKNOWN,
            sc_types.EDGE_ACCESS_VAR_POS_PERM >> ScAlias.RELATION_EDGE,
            ScKeynodes[CommonIdentifiers.RREL_LAST],
        )
        last_elem_templates = template_search(template)
        if last_elem_templates:
            last_elem_template = last_elem_templates[0]
            delete_elements(last_elem_template.get(ScAlias.RELATION_EDGE))  # Delete edge between rrel_last and edge
            return last_elem_template.get(ScAlias.ACCESS_EDGE)

        # Search unmarked last edge
        next_elem_result = search_role_relation_template(self._set_node, ScKeynodes[CommonIdentifiers.RREL_ONE])
        while True:
            next_edge = next_elem_result.get(ScAlias.ACCESS_EDGE)
            next_elem_result = self._search_next_element_template(next_edge)
            if next_elem_result is None:
                return next_edge

    def _create_next_edge(self, previous_edge: ScAddr, element: ScAddr) -> ScAddr:
        """Create edge to element and connect with previous edge"""
        template = ScTemplate()
        template.triple(
            self._set_node,
            sc_types.EDGE_ACCESS_VAR_POS_PERM >> ScAlias.ACCESS_EDGE,
            element,
        )
        template.triple_with_relation(
            previous_edge,
            sc_types.EDGE_D_COMMON_VAR,
            ScAlias.ACCESS_EDGE,
            sc_types.EDGE_ACCESS_VAR_POS_PERM,
            ScKeynodes[CommonIdentifiers.NREL_BASIC_SEQUENCE],
        )
        return template_generate(template).get(ScAlias.ACCESS_EDGE)

    @staticmethod
    def _mark_edge_with_rrel_last(last_edge: ScAddr) -> None:
        create_edge(sc_types.EDGE_ACCESS_CONST_POS_PERM, ScKeynodes[CommonIdentifiers.RREL_LAST], last_edge)

    def _search_next_element_template(self, cur_element_edge: ScAddr) -> Optional[ScTemplateResult]:
        templ = ScTemplate()
        templ.triple_with_relation(
            cur_element_edge,
            sc_types.EDGE_D_COMMON_VAR,
            sc_types.EDGE_ACCESS_VAR_POS_PERM >> ScAlias.ACCESS_EDGE,
            sc_types.EDGE_ACCESS_VAR_POS_PERM,
            ScKeynodes[CommonIdentifiers.NREL_BASIC_SEQUENCE],
        )
        templ.triple(self._set_node, ScAlias.ACCESS_EDGE, sc_types.UNKNOWN >> ScAlias.ELEMENT)
        search_results = template_search(templ)
        return search_results[0] if search_results else None



"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""

from typing import List, Optional, Union

from sc_client import client
from sc_client.constants import sc_types
from sc_client.constants.sc_types import ScType
from sc_client.models import ScAddr, ScConstruction, ScLinkContent, ScLinkContentType, ScTemplate, ScTemplateResult
from sc_client.models.sc_construction import ScLinkContentData

from sc_kpm.identifiers import CommonIdentifiers, ScAlias
from sc_kpm.sc_keynodes import Idtf, ScKeynodes


def create_nodes(*node_types: ScType) -> List[ScAddr]:
    construction = ScConstruction()
    for node_type in node_types:
        construction.create_node(node_type)
    return client.create_elements(construction)


def create_node(node_type: ScType) -> ScAddr:
    return create_nodes(node_type)[0]


def create_links(
    *contents: Union[str, int],
    content_type: ScLinkContentType = ScLinkContentType.STRING,
    link_type: ScType = sc_types.LINK_CONST,
) -> List[ScAddr]:
    construction = ScConstruction()
    for content in contents:
        link_content = ScLinkContent(content, content_type)
        construction.create_link(link_type, link_content)
    return client.create_elements(construction)


def create_link(
    content: Union[str, int],
    content_type: ScLinkContentType = ScLinkContentType.STRING,
    link_type: ScType = sc_types.LINK_CONST,
) -> ScAddr:
    return create_links(content, content_type=content_type, link_type=link_type)[0]


def create_edge(edge_type: ScType, src: ScAddr, trg: ScAddr) -> ScAddr:
    return create_edges(edge_type, src, trg)[0]


def create_edges(edge_type: ScType, src: ScAddr, *targets: ScAddr) -> List[ScAddr]:
    construction = ScConstruction()
    for trg in targets:
        construction.create_edge(edge_type, src, trg)
    return client.create_elements(construction)


def create_binary_relation(edge_type: ScType, src: ScAddr, trg: ScAddr, *relations: ScAddr) -> ScAddr:
    construction = ScConstruction()
    construction.create_edge(edge_type, src, trg, ScAlias.RELATION_EDGE)
    for relation in relations:
        construction.create_edge(sc_types.EDGE_ACCESS_CONST_POS_PERM, relation, ScAlias.RELATION_EDGE)
    return client.create_elements(construction)[0]


def create_role_relation(src: ScAddr, trg: ScAddr, *rrel_nodes: ScAddr) -> ScAddr:
    return create_binary_relation(sc_types.EDGE_ACCESS_CONST_POS_PERM, src, trg, *rrel_nodes)


def create_norole_relation(src: ScAddr, trg: ScAddr, *nrel_nodes: ScAddr) -> ScAddr:
    return create_binary_relation(sc_types.EDGE_D_COMMON_CONST, src, trg, *nrel_nodes)


def check_edge(edge_type: ScType, source: ScAddr, target: ScAddr) -> bool:
    return bool(get_edges(source, target, edge_type))


def get_edge(source: ScAddr, target: ScAddr, edge_type: ScType) -> ScAddr:
    edges = get_edges(source, target, edge_type)
    return edges[0] if edges else ScAddr(0)


def get_edges(source: ScAddr, target: ScAddr, *edge_types: ScType) -> List[ScAddr]:
    result_edges = []
    for edge_type in edge_types:
        templ = ScTemplate()
        templ.triple(source, edge_type, target)
        results = client.template_search(templ)
        result_edges.extend(result[1] for result in results)
    return result_edges


def get_system_idtf(addr: ScAddr) -> Idtf:
    nrel_system_idtf = ScKeynodes[CommonIdentifiers.NREL_SYSTEM_IDENTIFIER]

    templ = ScTemplate()
    templ.triple_with_relation(
        addr,
        sc_types.EDGE_D_COMMON_VAR,
        sc_types.LINK_VAR >> ScAlias.LINK,
        sc_types.EDGE_ACCESS_VAR_POS_PERM,
        nrel_system_idtf,
    )
    result = client.template_search(templ)
    if result:
        return get_link_content_data(result[0].get(ScAlias.LINK))
    return ""


def _search_relation_template(src: ScAddr, rel_node: ScAddr, rel_type: ScType) -> Optional[ScTemplateResult]:
    template = ScTemplate()
    template.triple_with_relation(
        src,
        rel_type >> ScAlias.ACCESS_EDGE,
        sc_types.UNKNOWN >> ScAlias.ELEMENT,
        sc_types.EDGE_ACCESS_VAR_POS_PERM,
        rel_node,
    )
    result = client.template_search(template)
    return result[0] if result else None


def search_role_relation_template(src: ScAddr, rrel_node: ScAddr) -> Optional[ScTemplateResult]:
    return _search_relation_template(src, rrel_node, sc_types.EDGE_ACCESS_VAR_POS_PERM)


def search_norole_relation_template(src: ScAddr, nrel_node: ScAddr) -> Optional[ScTemplateResult]:
    return _search_relation_template(src, nrel_node, sc_types.EDGE_D_COMMON_VAR)


def get_element_by_role_relation(src: ScAddr, rrel_node: ScAddr) -> ScAddr:
    search_result = search_role_relation_template(src, rrel_node)
    return search_result.get(ScAlias.ELEMENT) if search_result else ScAddr(0)


def get_element_by_norole_relation(src: ScAddr, nrel_node: ScAddr) -> ScAddr:
    search_result = search_norole_relation_template(src, nrel_node)
    return search_result.get(ScAlias.ELEMENT) if search_result else ScAddr(0)


def get_link_content_data(link: ScAddr) -> ScLinkContentData:
    content_part = client.get_link_content(link)
    return content_part[0].data


def delete_edges(source: ScAddr, target: ScAddr, *edge_types: ScType) -> bool:
    return client.delete_elements(*get_edges(source, target, *edge_types))
